from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Callable

from .consumer import Consumer, ConsumerGroup, ForwardResponse
from .logger import LoggerMixin
from .models import CloudEvent
from .settings import ServiceSettings
from .utils import generate_instance_id

if TYPE_CHECKING:
    from .broker import Broker
    from .types import TagMeta


class Service(LoggerMixin):
    """Logical group of consumers. Provides group (queue) name and handles versioning"""

    def __init__(
        self,
        name: str,
        broker: Broker,
        title: str | None = None,
        version: str = "0.1.0",
        description: str = "",
        tags_metadata: list[TagMeta] | None = None,
        instance_id_generator: Callable[[], str] | None = None,
        **context: Any,
    ):
        self.broker = broker
        self.name = name
        self.title = title or name.title()
        self.version = version
        self.description = description
        self.tags_metadata = tags_metadata or []
        self.id = (instance_id_generator or generate_instance_id)()
        self.consumer_group = ConsumerGroup()
        self.context = context
        # TODO: task gathering?
        self._tasks: list[asyncio.Task] = []

    def subscribe(
        self,
        topic: str,
        *,
        name: str | None = None,
        timeout: int = 120,
        dynamic: bool = False,
        forward_response: ForwardResponse | None = None,
        **options,
    ):
        return self.consumer_group.subscribe(
            topic=topic,
            name=name,
            timeout=timeout,
            dynamic=dynamic,
            forward_response=forward_response,
            **options,
        )

    def add_consumer(self, consumer: Consumer):
        self.consumer_group.add_consumer(consumer)

    def add_consumer_group(self, consumer_group: ConsumerGroup) -> None:
        self.consumer_group.add_consumer_group(consumer_group)

    async def publish(
        self,
        topic: str,
        data: Any | None = None,
        type_: type[CloudEvent] | str = "CloudEvent",
        **kwargs,
    ):
        kwargs.setdefault("source", self.name)
        return await self.broker.publish(topic, data, type_, **kwargs)

    @property
    def consumers(self):
        return self.consumer_group.consumers

    async def publish_event(self, message: CloudEvent, **kwargs):
        if not message.source:
            message.set_source(self.name)
        return await self.broker.publish_event(message, **kwargs)

    async def start(self):
        await self.broker.connect()
        await self.broker.dispatch_before("service_start", self)
        for consumer in self.consumers.values():
            asyncio.create_task(self.broker.start_consumer(self, consumer))
        await self.broker.dispatch_after("service_start", self)

    async def stop(self, *args, **kwargs):
        await self.broker.dispatch_before("service_stop", self)
        await self.broker.disconnect()
        await self.broker.dispatch_after("service_stop", self)

    def run(self, *args, **kwargs):
        from .runner import ServiceRunner

        runner = ServiceRunner([self])
        runner.run(*args, **kwargs)

    @classmethod
    def from_settings(cls, settings: ServiceSettings, **kwargs: Any) -> Service:
        kw = settings.dict(exclude={"broker_settings"})
        kw.update(**kwargs)
        broker = Broker.from_settings(settings.broker_settings)
        return cls(broker=broker, **kw)

    @classmethod
    def from_env(cls, **kwargs: Any) -> Service:
        return cls.from_settings(ServiceSettings(), **kwargs)



from typing import Any, Dict, List, Optional, Type

from pydantic import BaseModel as _BaseModel
from pydantic import Field

from eventiq.models import CloudEvent


class BaseModel(_BaseModel):
    class Config:
        allow_population_by_field_name = True


class Info(BaseModel):
    title: str
    version: str


class PublishInfo(BaseModel):
    topic: str
    event_type: Type[CloudEvent]
    kwargs: Dict[str, Any]


class PayloadRef(BaseModel):
    ref: str = Field(..., alias="$ref")


class Message(BaseModel):
    message_id: str = Field(..., alias="messageId")
    payload: PayloadRef
    content_type: str = Field(..., alias="contentType")
    description: Optional[str] = None
    tags: List[str] = []


class Operation(BaseModel):
    summary: Optional[str] = None
    message: Message


class ChannelItem(BaseModel):
    publish: Optional[Operation] = None
    subscribe: Optional[Operation] = None


class Server(BaseModel):
    url: Optional[str] = None
    protocol: str
    description: Optional[str] = None


class ExternalDocs(BaseModel):
    description: Optional[str] = None
    url: str


class Tag(BaseModel):
    name: str
    description: Optional[str] = None
    external_docs: Optional[str] = Field(None, alias="externalDocs")


class Components(BaseModel):
    schemas: Dict[str, Any]


class AsyncAPI(BaseModel):
    asyncapi: str = "2.5.0"
    info: Info
    servers: Dict[str, Server] = {}
    channels: Dict[str, ChannelItem] = {}
    default_content_type: str = Field("application/json", alias="defaultContentType")
    components: Components



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

import nats
from nats.aio.client import Client
from nats.aio.msg import Msg as NatsMsg
from nats.js import JetStreamContext

from eventiq.broker import Broker
from eventiq.exceptions import PublishError
from eventiq.utils.functools import retry

from ...message import Message
from .settings import JetStreamSettings, NatsSettings

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, Service


class NatsMessageProxy(Message[NatsMsg]):
    def __init__(self, message: NatsMsg):
        super().__init__(message)
        self._num_delivered = message.metadata.num_delivered


class NatsBroker(Broker[NatsMsg]):
    """
    :param url: Url to nats server(s)
    :param connection_options: additional connection options passed to nats.connect(...)
    :param auto_flush: auto flush messages on publish
    :param kwargs: options for base class
    """

    protocol = "nats"

    def __init__(
        self,
        *,
        url: str = "nats://localhost:4444",
        connection_options: dict[str, Any] | None = None,
        auto_flush: bool = True,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.url = url
        self.connection_options = connection_options or {}
        self._auto_flush = auto_flush
        self.client = Client()

    @property
    def message_proxy_class(self) -> type[Message]:
        return NatsMessageProxy

    def parse_incoming_message(self, message: NatsMsg) -> Any:
        return self.encoder.decode(message.data)

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        await self.client.subscribe(
            subject=consumer.topic,
            queue=f"{service.name}:{consumer.name}",
            cb=self.get_handler(service, consumer),
        )

    async def _disconnect(self) -> None:
        await self.client.close()

    async def flush(self):
        await self.client.flush()

    @retry(max_retries=3)
    async def _connect(self) -> None:
        self._nc = await self.client.connect(self.url, **self.connection_options)

    @retry(max_retries=3)
    async def _publish(self, message: CloudEvent, **kwargs) -> None:
        data = self.encoder.encode(message.dict())
        await self.client.publish(message.topic, data, **kwargs)
        if self._auto_flush:
            await self.client.flush()

    @property
    def is_connected(self) -> bool:
        return self.client.is_connected

    Settings = NatsSettings


class JetStreamBroker(NatsBroker):
    """
    NatsBroker with JetStream enabled
    :param prefetch_count: default number of messages to prefetch
    :param fetch_timeout: timeout for subscription pull
    :param jetstream_options: additional options passed to nc.jetstream(...)
    :param kwargs: all other options for base classes NatsBroker, Broker
    """

    Settings = JetStreamSettings

    def __init__(
        self,
        *,
        prefetch_count: int = 10,
        fetch_timeout: int = 10,
        jetstream_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.prefetch_count = prefetch_count
        self.fetch_timeout = fetch_timeout
        self.jetstream_options = jetstream_options or {}
        self.js = JetStreamContext(self.client, **self.jetstream_options)

    @retry(max_retries=3)
    async def _publish(
        self,
        message: CloudEvent,
        **kwargs,
    ) -> None:
        data = self.encoder.encode(message)
        headers = kwargs.get("headers", {})
        timeout = kwargs.get("timeout")
        stream = kwargs.get("stream")
        headers.setdefault("Content-Type", self.encoder.CONTENT_TYPE)
        try:
            await self.js.publish(
                subject=message.topic,
                payload=data,
                timeout=timeout,
                stream=stream,
                headers=headers,
            )
        except Exception as e:
            raise PublishError from e

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        durable = f"{service.name}:{consumer.name}"
        subscription = await self.js.pull_subscribe(
            subject=consumer.topic,
            durable=durable,
            config=consumer.options.get("config"),
        )
        handler = self.get_handler(service, consumer)
        batch = consumer.options.get("prefetch_count", self.prefetch_count)
        timeout = consumer.options.get("fetch_timeout", self.fetch_timeout)
        try:
            while not self._stopped:
                try:
                    messages = await subscription.fetch(batch=batch, timeout=timeout)
                    tasks = [asyncio.create_task(handler(message)) for message in messages]  # type: ignore
                    await asyncio.gather(*tasks, return_exceptions=True)
                except nats.errors.TimeoutError:
                    await asyncio.sleep(5)
        except Exception:
            self.logger.exception("Cancelling consumer")
        finally:
            if consumer.dynamic:
                await subscription.unsubscribe()

    async def _ack(self, message: Message) -> None:
        if not message._ackd:
            await message.ack()

    async def _nack(self, message: Message) -> None:
        if not message._ackd:
            await message.nak(delay=message.delay)



from enum import Enum, auto


class AutoEnum(str, Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name

    @staticmethod
    def auto():
        """
        Exposes `enum.auto()` to avoid requiring a second import to use `AutoEnum`
        """
        return auto()

    def __repr__(self) -> str:
        return f"{type(self).__name__}.{self.value}"
